(* Type entier, definition. Sans le signe *)
 
(* donne la longueur d'une liste sans les zeros derriere *)
let longueur liste=long_aux liste 0 0 where
     rec long_aux list k long=match list with
       |[]->k
       |a::q when a=0->long_aux q k (long+1)
       |a::q->long_aux q (long+1) (long+1);;


(* supprime les zeros en tete d'une liste *)
let rec purifie liste=match liste with
        |a::q when a=0->purifie q
        |a::q->liste
        |[]->[0];;
 
(* supprime les zeros en fin d'une liste *)
let purifie_par_la_fin liste=rev(aux liste [] []) where
       rec aux reste purifie_partiel sauv=match reste with
       |[]->purifie_partiel
       |a::suite when a=0->aux suite purifie_partiel (0::sauv)
       |a::suite->aux suite (a::sauv@purifie_partiel) [];;
 
(*addition de deux entiers longs *) 
let plus p a b=plus_aux a b 0 where
    rec plus_aux m n reste=match(m,n) with
    |([],[])->if reste=0 then [] else [reste]
    |([],u::q) when (u+reste)<p->(u+reste)::q
    |([],u::q)->((u+reste) mod p)::(plus_aux [] q ((u+reste)/p))
    |(u::q,[]) when (u+reste)<p->(u+reste)::q
    |(u::q,[])->((u+reste) mod p)::(plus_aux [] q ((u+reste)/p))
    |(u::q,v::s)->((u+v+reste) mod p)::(plus_aux q s ((u+v+reste)/p));;


(* soustraction de deux entiers longs, le premier tant plus grand que 
le deuxieme *)

let moins p a b=purifie_par_la_fin (moins_aux  a b 0) where 
         rec moins_aux m n reste=match(m,n) with
         |([],[])->[]
         |([u],[])->if (u-reste=0) then [] else [u-reste]
         |(u::q,[]) when (u-reste)>=0->(u-reste)::q
         |(u::q,[])->(u-reste+p)::(moins_aux q [] 1)
         |(u::q,v::s) when (u-v-reste)>=0->(u-v-reste)::(moins_aux q s 0)
         |(u::q,v::s)->(u-v-reste+p)::(moins_aux q s 1);;

(* transforme un entier long en integer *)
let int_of_entier p a=
     if (list_length a)>9 then failwith "nombre trop grand pour Caml"
     else subst_aux (rev a) 0 
        where rec subst_aux l n=match l with
       |[]->n
       |c::q->subst_aux q (n*p+c);;

(* transforme un int en entier long *)
let rec entier_of_int p a=match a with
         |0->[]
         |b->(b mod p)::(entier_of_int p (b/p));;

(* converti un string en entier long *)
let entier_of_string chaine=let long=string_length chaine in
          conv_aux [] 0
          where rec conv_aux n k=match k with
               |k when k=long->n
               |_->conv_aux (int_of_string (char_for_read (nth_char chaine k))::n) (k+1);; 

(* converti un entier long en chaine de caractres long *)
let string_of_entier n=aux n ""
       where rec aux l chaine=match l with
          |[]->chaine
          |a::q->aux q ((string_of_int a)^chaine);;


(* pour la multiplication d'un entier long par une puissance de p *)

let ee a n=aux a n where
    rec aux l k=match k with
        |0->l
        |k->aux (0::l) (k-1);; 

(* multiplication d'un entier (liste) long par un nombre de {0..9} *)
let mult p ent i=mult_aux ent 0 []
    where rec mult_aux  l rest accu=match l with
      |[]->if rest= 0 then rev accu
               else rev (rest::accu)
      |a::q->mult_aux q ((i*a+rest)/p) (((i*a+rest) mod p)::accu);;
 

(* multiplication naive de deux entiers longs *)
let fois p m n=aux1 m [0] 0 where
       rec aux1 a result k=match a with
          |[]->result
          |i::q->aux1 q (plus p result (ee (mult p n i) k)) (k+1);;

                       (* Multiplication rapide, plus rapide  partir d'une liste de taille environ 50 *)

(* exponentiation rapide d'integers*)
let rec puissance a=fun
        |0->1
        |k when (k mod 2)=0->let b=puissance a (k/2) in b*b
        |k->let b=puissance a (k/2) in a*b*b;;

(* prend comme argument deux listes quelconques et les remplies de 
zeros de facons  ce que elles aient toutes les deux une taille egale  la 
meme puissance de 2, et renvoie cette taille*)

let complete m n=let (a,b)=list_length m,list_length n in
                 let k=max a b in 
                 let q=puissance 2 (1+int_of_float (log (float_of_int 
k)/.log 2.)) in
                 (m@(ee [](q-a)),n@(ee [](q-b)),q);;



(* prend en argument 2 liste de longueur q une puissance de 2 et les 
decoupe en deux *)
let decoupe m n q=dec_aux (m,n,[],[],0)
               where rec dec_aux (lis1,lis2,nouv1,nouv2,k)=
                  if k=(q/2) then (lis1,rev nouv1,lis2,rev nouv2)
                  else match(lis1,lis2) with
                      |t::q,s::r->dec_aux (q,r,t::nouv1,s::nouv2,k+1)
                      |_->failwith "ERREUR! ERREUR! L'ORDINATEUR VA EXPLOSER DANS 5 SECONDES.";;
 
(* multiplicatiion rapide *)
let karatsuba p nombre1 nombre2=
       let(m,n,q)=complete nombre1 nombre2 in
           rev (purifie(rev(karatsuba_aux m n q)))
           where rec karatsuba_aux a b k=match k with
               |k when k<=12->fois p a b  (* 12 est la valeur pour laquelle karatsuba est le plus rapide *)
               |_->let(a1,a2,b1,b2)=decoupe a b k and
                              r=k/2   in
                  let s1=karatsuba_aux a1 b1 r in
                  let s2=karatsuba_aux (plus p a1 a2) (plus p b1 b2) r in
                  let s3=karatsuba_aux a2 b2 r in
                  let s4=moins p s2 (plus p s1 s3) in
                      plus p s3 (plus p (ee s4 r) (ee s1 r));;

(* Trouve la racine d'un poly
   p:base; polynome le poly; polynome1 la drive de polynome; g0 une solution modulo p et s0 l'inverse de polynomme1 g0*)
let racine p polynome polynome1 g0 s0 l=
  let g=ref g0 and s=ref s0 and modulo=ref p in
  for i=1 to int_of_float(log((float_of_int(l))/.log(2.))) do
    modulo:=!modulo* !modulo;
    g:=(!g - (polynome !g)* !s) mod (!modulo);
    s:=(2* !s- (polynome1 !g)* !s* !s) mod (!modulo);  
  done;
  modulo:=1;
  for i=1 to l do
    modulo:=p* !modulo;
  done;
  g:=(!g - (polynome !g)* !s) mod (!modulo);
  if !g>0 then entier_of_int p !g
  else entier_of_int p (!g+ !modulo);;
