#open "num";;

(* Type entier, definition. Sans le signe *)
type p_adique={Ent: num list;Val: int};;
type monome={Coeff:num;Degr:num};;
type polynome={Corps:monome list};;

let num_zero = num_of_int 0;;
let num_un = num_of_int 1;;

(* transforme un num en entier long *)
let rec entier_of_num p a=if a=num_zero then [] else
                          (mod_num a p)::(entier_of_num p (floor_num(a // p)));;

(* transforme un entier long en num *)
let num_of_entier p a=subst_aux (rev a) num_zero 
  where rec subst_aux l n=match l with
    |[]->n
    |c::q->subst_aux q (n*/p+/c);;

(* pour la multiplication d'un entier long par une puissance de p *)
let ee a n=aux a n where
    rec aux l k=match k with
        |0->l
        |k->aux (num_zero::l) (k-1);; 

let euclide f g=
  let r0=ref f and r1=ref g in
  let s0=ref num_un and s1=ref num_zero in
  let t0=ref num_zero and t1=ref num_un in 
  let q=ref num_zero in
  while !r1 <> num_zero do
    q:=floor_num (!r0 // !r1);
    let r2= mod_num !r0 !r1 in
    let s2= !s0 -/ !q */ !s1 in
    let t2= !t0 -/ !q */ !t1 in
    r0:= !r1;r1:=r2; 
    s0:= !s1;s1:=s2;
    t0:= !t1;t1:=t2;
  done;
  (!r0,!s0,!t0);;
  
let moins p x precision=
  let moins_un=ref num_zero and p'=p -/ num_un in
  for i=0 to precision do
    moins_un:= !moins_un +/ p' */ p **/ (num_of_int i)
  done;
  entier_of_num p (mod_num (!moins_un */ x) (p **/ (num_of_int precision)));;

let inverse p x precision=
  let (a,b,c)=euclide p (mod_num x p) in
  let g=ref (mod_num c p)  and modulo=ref p in
  for i=1 to int_of_float(log((float_of_int(precision))/.log(2.)))+1 do
    modulo:=!modulo */ !modulo;
    g:=mod_num ((num_of_int(2) */ !g -/ x */ !g */ !g)) (!modulo); 
  done;																																																																								  																																																																												  
  entier_of_num p !g;;
  
let eval_poly poly x=
  let rec eval l r=match l with
    |[]->r
    |p::q->eval q (r +/ p.Coeff */ x **/ p.Degr)
  in eval poly.Corps num_zero;;

let derive poly= 
  let rec der=function
    |[]->[]
    |{Coeff=a;Degr=b}::q->if b >/ num_zero then {Coeff=a */ b;Degr=b -/ num_un} :: der q else der q
  in {Corps=der poly.Corps};;  
  
(* Trouve la racine d'un poly
   p:base; polynome le poly; polynome1 la drive de polynome; g0 une solution modulo p*)  
let racine p poly g0 l=
  let poly'=derive poly in
  let (a,b,c)=euclide p (eval_poly poly' g0) in
  let g=ref g0 and s=ref (mod_num c p) and modulo=ref p in
  for i=1 to int_of_float(log((float_of_int(l))/.log(2.))) do
    modulo:=!modulo */ !modulo;
    g:=mod_num (!g -/ (eval_poly poly !g) */ !s) (!modulo);
    s:=mod_num (num_of_int(2) */ !s -/ (eval_poly poly' !g) */ !s */ !s) (!modulo);  
  done;
  entier_of_num p !g;;  


(* ******************************************** P-ADIQUE ************************************************************ *)

(* Les oprations de base (gre la valuation) *)
let zero={Ent=[num_zero];Val=0};;
let un={Ent=[num_un];Val=0};;

let compact n=
  let v=ref n.Val and l=ref n.Ent in
  while !l<>[] && hd(!l)=num_zero do
    l:=tl(!l);
    incr v
  done;
  if !l=[] then zero
  else {Ent= !l;Val= !v};;

let expand n=
  if n.Val > 0 then {Ent=ee n.Ent n.Val;Val=0}   
  else n;;

let entier p n=compact {Ent=entier_of_num p (num_of_int n);Val=0};;

let add_precision x n=
  if n>0 then {Ent=rev(ee (rev x.Ent) n);Val=x.Val}
  else x;;

let remove_precision x n=  
  if n>0 then 
  (let l=ref (rev x.Ent) and compt=ref n in
  while !l<>[] && !compt>0 do
    l:=tl(!l);
    decr(compt);
  done;
  if !l=[] then l:=[num_zero];
  {Ent=rev !l;Val= x.Val})
  else x;;
  
let normaliser p q=
  let m=p.Val and n=q.Val in
  let (p',q') =
    if n>m then p,{Ent=ee q.Ent (n-m);Val=m}
    else {Ent=ee p.Ent (m-n); Val=n},q
  in (p',q');;

let add p x y=
  let (x',y')=normaliser x y in
  compact {Ent=entier_of_num p ((num_of_entier p x'.Ent) +/ (num_of_entier p y'.Ent));Val=x'.Val};;
 
let multiply p x y=
  let (x',y')=normaliser x y in
  compact {Ent=entier_of_num p ((num_of_entier p x'.Ent) */ (num_of_entier p y'.Ent));Val=x'.Val+y'.Val};;

let opp p x precision=	
  compact {Ent=moins p (num_of_entier p x.Ent) precision;Val=x.Val};;

let minus p x y precision=
  add p x (opp p y precision);;

let invert p x precision=
  compact {Ent=inverse p (num_of_entier p x.Ent) precision;Val=0-x.Val};;
  
let convert t=
  let l=ref [] in
  for i=0 to vect_length t -1 do
    l:={Coeff=num_of_int(t.(i));Degr=num_of_int(i)}:: !l
  done;
  {Corps= !l};;

let root p poly s precision=
  compact {Ent=racine p poly s precision;Val=0};;